@@ -18,6 +18,7 @@
  * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
  */
 
+#include <glib/gi18n.h>
 #include "appearance.h"
 #include "appearance-desktop.h"
 #include "appearance-font.h"
@@ -29,12 +30,6 @@
 #include "activate-settings-daemon.h"
 #include "capplet-util.h"
 
-/* required for gnome_program_init(): */
-#include <libgnome/libgnome.h>
-#include <libgnomeui/gnome-ui-init.h>
-/* ---------------------------------- */
-#include <libgnomeui/gnome-thumbnail.h>
-
 static AppearanceData *
 init_appearance_data (int *argc, char ***argv)
 {
@@ -60,7 +55,7 @@ init_appearance_data (int *argc, char ***argv)
     data = g_new (AppearanceData, 1);
     data->client = gconf_client_get_default ();
     data->xml = ui;
-    data->thumb_factory = gnome_thumbnail_factory_new (GNOME_THUMBNAIL_SIZE_NORMAL);
+    data->thumb_factory = gnome_desktop_thumbnail_factory_new (GNOME_DESKTOP_THUMBNAIL_SIZE_NORMAL);
   }
 
   return data;
@@ -85,13 +80,39 @@ main_window_response (GtkWidget *widget,
     g_object_unref (data->client);
     g_object_unref (data->xml);
   }
+  else if (response_id == GTK_RESPONSE_HELP)
+  {
+      GtkNotebook *nb;
+      gint pindex;
+
+      nb = GTK_NOTEBOOK (glade_xml_get_widget (data->xml, "main_notebook"));
+      pindex = gtk_notebook_get_current_page (nb);
+
+      switch (pindex)
+      {
+        case 0: /* theme */
+          capplet_help (GTK_WINDOW (widget), "goscustdesk-12"); 
+          break;
+        case 1: /* background */
+          capplet_help (GTK_WINDOW (widget), "goscustdesk-7"); 
+          break;
+        case 2: /* fonts */
+          capplet_help (GTK_WINDOW (widget), "goscustdesk-38"); 
+          break;
+        case 3: /* interface */
+          capplet_help (GTK_WINDOW (widget), "goscustuserinter-2"); 
+          break;
+        default:
+          capplet_help (GTK_WINDOW (widget), "prefs-look-and-feel"); 
+          break;
+       }
+  }
 }
 
 int
 main (int argc, char **argv)
 {
   AppearanceData *data;
-  GnomeProgram *program;
   GtkWidget *w;
 
   gchar *install_filename = NULL;
@@ -132,19 +153,14 @@ main (int argc, char **argv)
   option_context = g_option_context_new (NULL);
   g_option_context_add_main_entries (option_context, option_entries, GETTEXT_PACKAGE);
 
-  /* this appears to be required for gnome_wm_manager_init (), which is called
-   * inside gnome_meta_theme_set ();
-   * TODO: try to remove if possible
-   */
-  program = gnome_program_init ("appearance", VERSION,
-        LIBGNOMEUI_MODULE, argc, argv,
-        GNOME_PARAM_APP_DATADIR, GNOMECC_DATA_DIR,
-        GNOME_PARAM_GOPTION_CONTEXT, option_context,
-        NULL);
+  capplet_init (option_context, &argc, &argv);
 
-  if (install_filename != NULL)
-     gnome_theme_install_from_uri (install_filename, NULL);
-  g_free (install_filename);
+  if (install_filename != NULL) {
+    GFile *inst = g_file_new_for_commandline_arg (install_filename);
+    g_free (install_filename);
+    gnome_theme_install (inst, NULL);
+    g_object_unref (inst);
+  }
 
   /* init tabs */
   themes_init (data);
@@ -190,7 +206,6 @@ main (int argc, char **argv)
 
   /* free stuff */
   g_free (data);
-  g_object_unref (program);
 
   return 0;
 }
