@@ -27,7 +27,9 @@
 #include <glib/gi18n.h>
 #include <gio/gio.h>
 #include <glib/gstdio.h>
+#include <unistd.h>
 
+#include "capplet-util.h"
 #include "file-transfer-dialog.h"
 #include "theme-installer.h"
 #include "theme-util.h"
@@ -57,7 +59,7 @@ cleanup_tmp_dir (GIOSchedulerJob *job,
 	GFile *directory;
 
 	directory = g_file_new_for_path (tmp_dir);
-	file_delete_recursive (directory, NULL);
+	capplet_file_delete_recursive (directory, NULL);
 	g_object_unref (directory);
 
 	return FALSE;
@@ -248,7 +250,7 @@ invalid_theme_dialog (const gchar *filename, gboolean maybe_theme_engine)
 	const gchar *engine = _("\"%s\" does not appear to be a valid theme. It may be a theme engine which you need to compile.");
 
 	dialog = gtk_message_dialog_new (NULL, GTK_DIALOG_MODAL,
-			GTK_MESSAGE_ERROR, GTK_BUTTONS_OK, primary);
+			GTK_MESSAGE_ERROR, GTK_BUTTONS_OK, "%s", primary);
 	if (maybe_theme_engine)
 		gtk_message_dialog_format_secondary_text (
 				GTK_MESSAGE_DIALOG (dialog), engine, filename);
@@ -356,9 +358,10 @@ gnome_theme_install_real (gint filetype, const gchar *tmp_dir, const gchar *them
 					GTK_DIALOG_MODAL,
 					GTK_MESSAGE_ERROR,
 					GTK_BUTTONS_OK,
+					"%s",
 					str);
 		gtk_message_dialog_format_secondary_text (GTK_MESSAGE_DIALOG (dialog),
-							  error->message);
+							  "%s", error->message);
 
 		g_free (str);
 		g_error_free (error);
@@ -393,7 +396,7 @@ gnome_theme_install_real (gint filetype, const gchar *tmp_dir, const gchar *them
 				user_message = g_strdup_printf ("<span weight=\"bold\" size=\"larger\">%s</span>", str);
 				g_free (str);
 
-				dialog = gtk_message_dialog_new_with_markup (NULL, GTK_DIALOG_MODAL, GTK_MESSAGE_INFO, GTK_BUTTONS_NONE, user_message);
+				dialog = gtk_message_dialog_new_with_markup (NULL, GTK_DIALOG_MODAL, GTK_MESSAGE_INFO, GTK_BUTTONS_NONE, "%s", user_message);
 
 				gtk_message_dialog_format_secondary_text (GTK_MESSAGE_DIALOG (dialog), _("Would you like to apply it now, or keep your current theme?"));
 
@@ -443,6 +446,7 @@ gnome_theme_install_real (gint filetype, const gchar *tmp_dir, const gchar *them
 				       GTK_DIALOG_MODAL,
 				       GTK_MESSAGE_INFO,
 				       GTK_BUTTONS_OK,
+				       "%s",
 				       user_message);
 				gtk_dialog_run (GTK_DIALOG (dialog));
 			}
@@ -578,16 +582,14 @@ transfer_done_cb (GtkWidget *dlg, gchar *path)
 }
 
 void
-gnome_theme_install_from_uri (const gchar *filename, GtkWindow *parent)
+gnome_theme_install (GFile *file, GtkWindow *parent)
 {
 	GtkWidget *dialog;
 	gchar *path, *base;
 	GList *src, *target;
-	gchar *temppath;
 	const gchar *template;
-	int cmp;
 
-	if (filename == NULL || strcmp (filename, "") == 0) {
+	if (file == NULL) {
 		dialog = gtk_message_dialog_new (NULL,
 						 GTK_DIALOG_MODAL,
 						 GTK_MESSAGE_ERROR,
@@ -598,9 +600,10 @@ gnome_theme_install_from_uri (const gchar *filename, GtkWindow *parent)
 		return;
 	}
 
-	/* see if someone dropped a directory */
-	if (g_file_test (filename, G_FILE_TEST_IS_DIR))	{
-		transfer_done_cb (NULL, g_strdup (filename));
+	/* see if someone dropped a local directory */
+	if (g_file_is_native (file) &&
+	    g_file_query_file_type (file, G_FILE_QUERY_INFO_NONE, NULL) == G_FILE_TYPE_DIRECTORY) {
+		transfer_done_cb (NULL, g_file_get_path (file));
 		return;
 	}
 
@@ -619,28 +622,7 @@ gnome_theme_install_from_uri (const gchar *filename, GtkWindow *parent)
 		return;
 	}
 
-	/* To avoid the copy of /root/.themes to /root/.themes/.themes
-	 * which causes an infinite loop. The user asks to transfer the all
-	 * contents of a folder, to a folder under itself. So ignore the
-	 * situation.
-	 */
-	temppath = g_build_filename (filename, ".themes", NULL);
-	cmp = strcmp (temppath, path);
-	g_free (path);
-	g_free (temppath);
-
-	if (cmp == 0) {
-		dialog = gtk_message_dialog_new (NULL,
-						 GTK_DIALOG_MODAL,
-						 GTK_MESSAGE_ERROR,
-						 GTK_BUTTONS_OK,
-						 _("%s is the path where the theme files will be installed. This can not be selected as the source location"), filename);
-		gtk_dialog_run (GTK_DIALOG (dialog));
-		gtk_widget_destroy (dialog);
-		return;
-	}
-
-	base = g_path_get_basename (filename);
+	base = g_file_get_basename (file);
 
 	if (g_str_has_suffix (base, ".tar.gz") || g_str_has_suffix (base, ".tgz") || g_str_has_suffix (base, ".gtp"))
 		template = "gnome-theme-%d.gtp";
@@ -653,7 +635,7 @@ gnome_theme_install_from_uri (const gchar *filename, GtkWindow *parent)
 	}
 	g_free (base);
 
-	src = g_list_append (NULL, g_strdup (filename));
+	src = g_list_append (NULL, g_object_ref (file));
 
 	path = NULL;
 	do {
@@ -666,7 +648,7 @@ gnome_theme_install_from_uri (const gchar *filename, GtkWindow *parent)
 	} while (g_file_test (path, G_FILE_TEST_EXISTS));
 
 
-	target = g_list_append (NULL, path);
+	target = g_list_append (NULL, g_file_new_for_path (path));
 
 	dialog = file_transfer_dialog_new_with_parent (parent);
 	g_signal_connect (dialog, "cancel",
@@ -680,9 +662,10 @@ gnome_theme_install_from_uri (const gchar *filename, GtkWindow *parent)
 					 G_PRIORITY_DEFAULT);
 	gtk_widget_show (dialog);
 
-	g_list_foreach (src, (GFunc) g_free, NULL);
-	/* don't free the target item since we're using that for the signals */
+	/* don't free the path since we're using that for the signals */
+	g_list_foreach (src, (GFunc) g_object_unref, NULL);
 	g_list_free (src);
+	g_list_foreach (target, (GFunc) g_object_unref, NULL);
 	g_list_free (target);
 }
 
@@ -692,7 +675,6 @@ gnome_theme_installer_run (GtkWindow *parent, const gchar *filename)
 	static gboolean running_theme_install = FALSE;
 	static gchar old_folder[512] = "";
 	GtkWidget *dialog;
-	gchar *filename_selected, *folder;
 	GtkFileFilter *filter;
 
 	if (running_theme_install)
@@ -723,7 +705,9 @@ gnome_theme_installer_run (GtkWindow *parent, const gchar *filename)
 
 	if (gtk_dialog_run (GTK_DIALOG (dialog)) == GTK_RESPONSE_ACCEPT)
 	{
-		filename_selected = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER (dialog));
+		gchar *uri_selected, *folder;
+
+		uri_selected = gtk_file_chooser_get_uri (GTK_FILE_CHOOSER (dialog));
 
 		folder = gtk_file_chooser_get_current_folder (GTK_FILE_CHOOSER (dialog));
 		g_strlcpy (old_folder, folder, 255);
@@ -731,8 +715,13 @@ gnome_theme_installer_run (GtkWindow *parent, const gchar *filename)
 
 		gtk_widget_destroy (dialog);
 
-		gnome_theme_install_from_uri (filename_selected, parent);
-		g_free (filename_selected);
+		if (uri_selected != NULL) {
+			GFile *file = g_file_new_for_uri (uri_selected);
+			g_free (uri_selected);
+
+			gnome_theme_install (file, parent);
+			g_object_unref (file);
+		}
 	}
 	else
 		gtk_widget_destroy (dialog);
