@@ -31,6 +31,8 @@
 #include <sys/types.h>
 #include <sys/stat.h>
 #include <unistd.h>
+#include <glib/gi18n.h>
+#include <stdlib.h>
 
 #include "capplet-util.h"
 
@@ -65,20 +67,29 @@ capplet_error_dialog (GtkWindow *parent, char const *msg, GError *err)
  * Havoc happy way.
  **/
 void
-capplet_help (GtkWindow *parent, char const *helpfile, char const *section)
+capplet_help (GtkWindow *parent, char const *section)
 {
 	GError *error = NULL;
+	char *uri;
+	GdkScreen *screen;
 
-	g_return_if_fail (helpfile != NULL);
 	g_return_if_fail (section != NULL);
 
-	gnome_help_display_desktop (NULL,
-		"user-guide",
-		helpfile, section, &error);
-	if (error != NULL)
-		capplet_error_dialog (parent, 
+	if (!parent)
+		screen = gdk_screen_get_default ();
+	else
+		screen = gtk_widget_get_screen (GTK_WIDGET (parent));
+
+	uri = g_strdup_printf ("ghelp:user-guide#%s", section);
+
+	if (!gtk_show_uri (screen, uri, gtk_get_current_event_time (), &error)) {
+		capplet_error_dialog (
+			parent,
 			_("There was an error displaying help: %s"),
 			error);
+	}
+
+	g_free (uri);
 }
 
 /**
@@ -166,3 +177,30 @@ capplet_file_delete_recursive (GFile *file, GError **error)
 		return g_file_delete (file, NULL, error);
 }
 
+void
+capplet_init (GOptionContext *context,
+	      int *argc,
+	      char ***argv)
+{
+	GError *err = NULL;
+
+#ifdef ENABLE_NLS
+	bindtextdomain (GETTEXT_PACKAGE, GNOMELOCALEDIR);
+	bind_textdomain_codeset (GETTEXT_PACKAGE, "UTF-8");
+	textdomain (GETTEXT_PACKAGE);
+#endif
+
+	if (context) {
+#if GLIB_CHECK_VERSION (2, 12, 0)
+		g_option_context_set_translation_domain (context, GETTEXT_PACKAGE);
+#endif
+		g_option_context_add_group (context, gtk_get_option_group (TRUE));
+
+		if (!g_option_context_parse (context, argc, argv, &err)) {
+			g_printerr ("%s\n", err->message);
+			exit (1);
+		}
+	}
+
+	gtk_init (argc, argv);
+}
