@@ -5,12 +5,20 @@
 #endif
 
 #include <string.h>
+#include <stdio.h>
+#include <unistd.h>
+#include <time.h>
+#include <sys/stat.h>
+#include <sys/resource.h>
+#include <sys/time.h>
 #include <gconf/gconf-value.h>
 #include <libgnome/gnome-url.h>
 
 #define DESKTOP_ITEM_TERMINAL_EMULATOR_FLAG "TerminalEmulator"
 #define ALTERNATE_DOCPATH_KEY               "DocPath"
 
+static FILE *checkpoint_file;
+
 gboolean
 libslab_gtk_image_set_by_id (GtkImage *image, const gchar *id)
 {
@@ -562,3 +570,126 @@ libslab_spawn_command (const gchar *cmd)
 
 	g_strfreev (argv);
 }
+
+static guint thumbnail_factory_idle_id;
+static GnomeThumbnailFactory *thumbnail_factory;
+
+static void
+create_thumbnail_factory (void)
+{
+	/* The thumbnail_factory may already have been created by an applet
+	 * instance that was launched before the current one.
+	 */
+	if (thumbnail_factory != NULL)
+		return;
+
+	libslab_checkpoint ("create_thumbnail_factory(): start");
+
+	thumbnail_factory = gnome_thumbnail_factory_new (GNOME_THUMBNAIL_SIZE_NORMAL);
+
+	libslab_checkpoint ("create_thumbnail_factory(): end");
+}
+
+static gboolean
+init_thumbnail_factory_idle_cb (gpointer data)
+{
+	create_thumbnail_factory ();
+	thumbnail_factory_idle_id = 0;
+	return FALSE;
+}
+
+void
+libslab_thumbnail_factory_preinit (void)
+{
+	thumbnail_factory_idle_id = g_idle_add (init_thumbnail_factory_idle_cb, NULL);
+}
+
+GnomeThumbnailFactory *
+libslab_thumbnail_factory_get (void)
+{
+	if (thumbnail_factory_idle_id != 0) {
+		g_source_remove (thumbnail_factory_idle_id);
+		thumbnail_factory_idle_id = 0;
+
+		create_thumbnail_factory ();
+	}
+
+	g_assert (thumbnail_factory != NULL);
+	return thumbnail_factory;
+}
+
+void
+libslab_checkpoint_init (const char *checkpoint_config_file_basename,
+			 const char *checkpoint_file_basename)
+{
+	char *filename;
+	struct stat st;
+	int result;
+	time_t t;
+	struct tm tm;
+	char *checkpoint_full_basename;
+
+	g_return_if_fail (checkpoint_config_file_basename != NULL);
+	g_return_if_fail (checkpoint_file_basename != NULL);
+
+	filename = g_build_filename (g_get_home_dir (), checkpoint_config_file_basename, NULL);
+
+	result = stat (filename, &st);
+	g_free (filename);
+
+	if (result != 0)
+		return;
+
+	t = time (NULL);
+	tm = *localtime (&t);
+
+	checkpoint_full_basename = g_strdup_printf ("%s-%04d-%02d-%02d-%02d-%02d-%02d.checkpoint",
+						    checkpoint_file_basename,
+						    tm.tm_year + 1900,
+						    tm.tm_mon + 1,
+						    tm.tm_mday,
+						    tm.tm_hour,
+						    tm.tm_min,
+						    tm.tm_sec);
+
+	filename = g_build_filename (g_get_home_dir (), checkpoint_full_basename, NULL);
+	g_free (checkpoint_full_basename);
+
+	checkpoint_file = fopen (filename, "w");
+	g_free (filename);
+}
+
+void
+libslab_checkpoint (const char *format, ...)
+{
+	va_list args;
+	struct timeval tv;
+	struct tm tm;
+	struct rusage rusage;
+
+	if (!checkpoint_file)
+		return;
+
+	gettimeofday (&tv, NULL);
+	tm = *localtime (&tv.tv_sec);
+
+	getrusage (RUSAGE_SELF, &rusage);
+
+	fprintf (checkpoint_file,
+		 "%02d:%02d:%02d.%04d (user:%d.%04d, sys:%d.%04d) - ",
+		 (int) tm.tm_hour,
+		 (int) tm.tm_min,
+		 (int) tm.tm_sec,
+		 (int) (tv.tv_usec / 100),
+		 (int) rusage.ru_utime.tv_sec,
+		 (int) (rusage.ru_utime.tv_usec / 100),
+		 (int) rusage.ru_stime.tv_sec,
+		 (int) (rusage.ru_stime.tv_usec / 100));
+
+	va_start (args, format);
+	vfprintf (checkpoint_file, format, args);
+	va_end (args);
+
+	fputs ("\n", checkpoint_file);
+	fflush (checkpoint_file);
+}
